open Sql
open ExtList
open Value
module DbQ = Query.Simple

let (^/) a b = a ^ "/" ^ b

let get dbh map query =
  PGOCaml.prepare dbh ~query:("SELECT " ^ query) ();
  let x = map (Option.get (List.hd (List.hd (PGOCaml.execute dbh ~params:[] ())))) in
    PGOCaml.close_statement dbh ();
    (x, )

let nextval dbh sequence =
  get dbh PGOCaml.int32_of_string ("NEXTVAL('" ^ sequence ^ "')")

let current_timestamp dbh = 
  get 
    dbh
    (fun v -> (PGOCaml.timestamp_of_string v, CalendarLib.Time_Zone.Local)) 
    "current_timestamp"

let create_document dbh ?(title=Value.string "") () =
  DbQ.query dbh <:insert< 
    $Rel.document_tbl$ := {
      document  = $int32:nextval dbh "document_document_seq"$;
      created   = $timestamptz:current_timestamp dbh$;
      title	= $title$;
      origin	= 0l
    }
  >>;
  (* for compatibility with the rest of the types *)
  Int64.to_int32 (PGOCaml.serial dbh "document_document_seq")

let create_page dbh =
  let id = nextval dbh "page_page_seq" in
    DbQ.query dbh <:insert<
      $Rel.page_tbl$ := {
	page	= $int32:id$;
	a	= 0;
	b	= 0;
	blank	= false
      }
    >>;
    id

let create_paper dbh document a b index =
  DbQ.query dbh <:insert<
    $Rel.paper_tbl$ := {
      a		= $int32:a$;
      b		= $int32:b$;
      document	= $int32:document$;
      index     = $int32:index$;
    }
  >>;
  DbQ.query dbh <:update<
    row in $Rel.page_tbl$ := {
      a         = $int32:a$;
      b         = $int32:b$
    } | row.page = $int32:a$ || row.page = $int32:b$
  >>

let all table = <:select< r | r in $table$ >>

let query_all dbh table = DbQ.query dbh (all table)

let required_formats =
  let formats = <:view< r | r in $Rel.require_document_format$ >> in
  let page_format = 
    <:view< r | 
	    r in $Rel.document_format$; f in $formats$;
            f.document_format = r.document_format >> in 
  let required  =
    <:view< group {page_format = k} by {k = r.require} |
            r in $page_format$ >>
  in
    required

let process_incoming dbh incoming_dir page_fs =
  let files = List.sort (FileUtils.directory_contents incoming_dir) in
    PGOCaml.begin_work dbh;
    let document = create_document dbh () in
      Printf.printf "Created document %ld\n%!" document;
      let pages, remaining_action =
	List.fold_left
	  (fun (pages, (page_action, index)) file -> 
	     let page = create_page dbh in
	     let next_action, next_index = match page_action with
	       | `A_side -> 
		   (* the action will be disregarded, if there
		      is a next page to be processed *)
		   `B_side 
		     ((fun () -> create_paper dbh document page page index),
		      page), index
	       | `B_side (_action, a_page) ->
		   create_paper dbh document a_page page index;
		   `A_side, Int32.succ index
	     in
	       Unix.link (incoming_dir ^/ file) (page_fs page);
	       Printf.printf "file: %s\n%!" file;
	       page::pages, (next_action, next_index)
	  )
	  ([], (`A_side, 1l))
	  files
      in
	( match remaining_action with
	    | `B_side (action, _), _ -> action ()
	    | _ -> () );
	PGOCaml.commit dbh;
	pages

let process_pages dbh page_name document_name pages =
  let formats = query_all dbh required_formats in
    List.iter
      (fun page -> ()
      )
      pages

let main () =
  let dbh = PGOCaml.connect () in
  let page_fs ?format n = 
    "pages" ^/ Int32.to_string n ^ 
      match format with
	| None -> ""
	| Some format -> "." ^ format
  in
  let document_fs n = "documents" ^/ Int32.to_string n in
  let pages = process_incoming dbh "incoming" page_fs in
    process_pages dbh document_fs pages
    
let _ = main ()
